/* 
 *  osc2UefC.cpp
 *  uefA Christof Eberspaecher
 *  uefA ohne Veraenderung laeuft
 *  uefB laeuft 11.02.24 11:10
 *  uefC.ino auf Arduino
 *  uefC.cpp 12.02.24 for Daisy Seed
 *  Status: laeuft mit Tasks 
 *  uefC.cpp 13.02.24 with Block Files in QSPI FLASH
 */
  
#include "daisy_seed.h"
#include "daisysp.h"

// Use the daisy namespace to prevent having to type
// daisy:: before all libdaisy functions
using namespace daisy;

// Declare a DaisySeed object called hardware
DaisySeed hw;

daisysp::Oscillator osc;
CpuLoadMeter loadMeter;

//                         ampli, freq, bufAdr 
volatile float daisyControls[100]= { 0.1, 220.0, };
volatile float daisyAudio[48]= { 0, }; // mono buffer to be filled by forth
void calledByForth(); // can be called by Forth

#include "ueForthC.cpp"




void AudioCallback(AudioHandle::InputBuffer in, 
                AudioHandle::OutputBuffer out, 
                size_t size) 
{
    float osc_out;
    loadMeter.OnBlockStart();

    osc.SetFreq(daisyControls[1]);
    osc.SetAmp(daisyControls[0]);
    daisyControls[50]= loadMeter.GetAvgCpuLoad();

    InterruptForth(); // work the interrupt routine defined in forth
    //Fill the block with samples
    for(size_t i = 0; i < size; i++)
    {
        //get the next oscillator sample
        osc_out = osc.Process();

        //Set the left and right outputs
        out[0][i] = 0.5*(osc_out+daisyAudio[i]);
        out[1][i] = 0.5*(osc_out+daisyAudio[i]);
    }

    loadMeter.OnBlockEnd();
}

void calledByForth() // do something triggered by Forth
{
  daisyControls[1] *= 2; // double frequency
}

void AudioStart() {
    // hw.SetAudioBlockSize(4);

    //How many samples we'll output per second
    float samplerate = hw.AudioSampleRate();

    //Set up oscillator
    osc.Init(samplerate);
    osc.SetWaveform(osc.WAVE_SIN);
    // initialize the load meter so that it knows what time is available for the processing:
    loadMeter.Init(hw.AudioSampleRate(), hw.AudioBlockSize());
    
    //Start calling the audio callback
    hw.StartAudio(AudioCallback);
}


#define hSize (48*1024)
cell_t heap[hSize/4];

void setupForth() {
  /*
  cell_t fh = heap_caps_get_free_size(MALLOC_CAP_INTERNAL);
  cell_t hc = heap_caps_get_largest_free_block(MALLOC_CAP_INTERNAL);
  if (fh - hc < MINIMUM_FREE_SYSTEM_HEAP) {
    hc = fh - MINIMUM_FREE_SYSTEM_HEAP;
  }
  */
  cell_t hc= hSize;
  forth_init(0, 0, heap, hc, boot, sizeof(boot));
  //Serial.println(LED_BUILTIN);
  hw.PrintLine("Blocks: %x", qspi_buffer+BL_OFFSET);
  hw.PrintLine("End of Setup");
 
}

void runForth() { // will not return
  g_sys.rp = forth_run(g_sys.rp);
}

int main()
{

	// Configure and Initialize the Daisy Seed
   // These are separate to allow reconfiguration of any of the internal
   // components before initialization.
   hw.Configure();
   hw.Init();
   hw.StartLog(true); // aktiviert USB-serial
	System::Delay(500);
	hw.usb_handle.SetReceiveCallback(UsbCallback,
                                     UsbHandle::UsbPeriph::FS_INTERNAL);
									 
	hw.PrintLine(" ");
    hw.Print(__FILE__);
	hw.PrintLine(__DATE__);
	//testUsb();
	//System::Delay(500);
   
   setupForth();
   AudioStart();
   runForth(); // will not return
}
